#include "NFC.h"
#include <SPI.h>
#include <PubSubClient.h>
#include "helpers.h"

NFC::NFC(int pin_sda, int pin_scl)
{
    Serial.println("NFC: Start");

    rfid = new Adafruit_PN532(PN532_IRQ, PN532_RESET);

    rfid->begin();

    uint32_t versiondata = rfid->getFirmwareVersion();
    if (!versiondata)
    {
        Serial.print("Didn't find PN53x board");
        while (1)
        {
            yield(); // halt
        }
    }

    // Got ok data, print it out!
    Serial.print("NFC: Found chip PN5");
    Serial.println((versiondata >> 24) & 0xFF, HEX);
    Serial.print("NFC: Firmware ver. ");
    Serial.print((versiondata >> 16) & 0xFF, DEC);
    Serial.print('.');
    Serial.println((versiondata >> 8) & 0xFF, DEC);
}

String NFC::getUID()
{
    String uidStr;
    for (int i = 0; i < uidLength; i++)
    {
        if (uid[i] < 0x10)
        {
            uidStr += "0";
        }
        uidStr += String(uid[i], HEX);
    }

    return uidStr;
}

bool NFC::hasCardSelected()
{
    return cardSelected;
}

bool NFC::testNFC()
{
    // return rfid->PCD_PerformSelfTest();
}

bool NFC::checkforCard()
{
    // Attempt to read a card (timeout in milliseconds, adjust as needed)
    if (!rfid->readPassiveTargetID(PN532_MIFARE_ISO14443A, uid, &uidLength, 1000))
    {
        Serial.println("No card found within the timeout period");
        // No card found within the timeout period
        return false;
    }

    // At this point, a card has been detected and its UID is in 'uid' with length 'uidLength'
    // If you expect a DESFire card, you might assume its UID is 7 bytes.
    // (Alternatively, if you extend the library to capture the SAK, you could check that here.)
    if (uidLength == 7)
    {
        Serial.println("DESFire card detected");
        Serial.print("UID: ");
        printbytes(uid, uidLength);

        return true;
    }

    // If we're getting 4 bytes, we might be scanning a MIFARE Classic or another ISO14443A card.
    if (uidLength == 4)
    {
        Serial.println("MIFARE Classic or another ISO14443A card detected");
        Serial.print("UID: ");
        printbytes(uid, uidLength);

        return true;
    }

    // The card detected does not match the expected DESFireor MIFARE Classic UID length
    Serial.println("No card detected");

    return false;
}

bool NFC::connecttoCard()
{
    // --- RATS (Request for ATS) ---
    // RATS command format: [0xE0, param]
    // The parameter byte encodes the Frame Size and other options.
    // Here we use 0x80 as an example parameter (adjust as needed for your card).
    uint8_t ratsCmd[] = {0xE0, 0x80};
    uint8_t atsResponse[16]; // Buffer to hold the ATS response
    uint8_t atsResponseLength = sizeof(atsResponse);

    // Send the RATS command via inDataExchange.
    // inDataExchange() returns a negative value on error.
    int result = rfid->inDataExchange(ratsCmd, sizeof(ratsCmd), atsResponse, &atsResponseLength);
    if (result < 0)
    {
        Serial.println(F("Failed ATS"));
        // PN532 does not offer a dedicated halt command,
        // so you may simply return false to abort communication.
        return false;
    }

    // (Optional) Process the ATS response here if needed.
    // The ATS provides card-specific information, such as supported protocols and frame sizes.

    // --- PPS (Protocol and Parameter Selection) ---
    // PPS command format (example): [0xD0, PPS0, PPS1]
    // In this example, PPS0 is set to 0x11 and PPS1 to 0x00.
    // Adjust the values according to your protocol requirements.
    uint8_t ppsCmd[] = {0xD0, 0x11, 0x00};
    uint8_t ppsResponse[16];
    uint8_t ppsResponseLength = sizeof(ppsResponse);

    result = rfid->inDataExchange(ppsCmd, sizeof(ppsCmd), ppsResponse, &ppsResponseLength);
    if (result < 0)
    {
        Serial.println(F("Failed PPS"));
        return false;
    }

    // If both exchanges were successful, mark the card as selected.
    cardSelected = true;

    // Save any card-specific data if needed here.
    // Note: Unlike MFRC522, Adafruit_PN532 does not store the UID in a public member.
    // We save the UID when we detect the card.
    pcb = 0x0A;
    return true;
}

bool NFC::disconnectCard()
{
    // ISO14443A HALT command: 0x50, 0x00, followed by the two-byte CRC.
    // You can either compute the CRC or, if your setup automatically appends it,
    // simply send the command without CRC bytes.
    // Here, we assume you have to include the CRC.
    // Note: The correct CRC for [0x50, 0x00] is usually 0x00, 0x00 if the PN532
    // is set to auto-calculate CRC, or you can compute it if needed.

    // Option 1: If your PN532 is configured to append the CRC automatically,
    // you can send just the two-byte command.
    uint8_t haltCmd[] = {0x50, 0x00};
    uint8_t response[8];
    uint8_t responseLength = sizeof(response);

    int status = rfid->inDataExchange(haltCmd, sizeof(haltCmd), response, &responseLength);
    if (status < 0)
    {
        Serial.println(F("Failed to send HALT command"));
        Serial.println(status);
        return false;
    }

    // Option 2: If you need to send a complete command with CRC, you must
    // calculate the CRC bytes for [0x50, 0x00]. For many ISO14443A cards, the CRC
    // may be 0x00, 0x00 when using a PN532 in auto-CRC mode, so this might work:
    //
    // uint8_t haltCmd[] = { 0x50, 0x00, 0x00, 0x00 };
    // int status = rfid->inDataExchange(haltCmd, sizeof(haltCmd), response, &responseLength);
    // if (status < 0) { ... }
    //
    // Uncomment the above block and comment out the Option 1 block if needed.

    return true;
}

bool NFC::testCard()
{
    return true;
}
